package CXC::ACIS::Command::Block;

use strict;
use warnings;

our $VERSION = '0.02';

use CXC::ACIS::Command::Constants qw[ %CMD ];
use CXC::ACIS::Command::Types -types;

use Try::Tiny;

use Types::XSD qw[ PositiveInteger ];
use Type::Params qw[ validate ];
use Types::Standard qw[ Enum ];

use Carp;

use Exporter qw[ import ];

our @EXPORT_OK = qw[ loadBlock ];

## no critic (ProhibitAccessOfPrivateData)

#<<< no perltidy

our %decode = (

    commandHeader => [
        commandLength     => { value => [  0, 16 ], fmt => '%d', isa => PositiveInteger },
        commandIdentifier => { value => [ 16, 16 ], fmt => '%d' 			},
        commandOpcode     => { value => [ 32, 16 ], fmt => '%d' 			},
    ],


    addBadColumn => [
        columns => {
            loop => { offset => 48, nbits => 16 },
            over => [
                ccdId 	  => { value => [ 0,  4 ], fmt => '%d', isa => CcdId 	 },
                ccdColumn => { value => [ 4, 10 ], fmt => '%d', isa => CcdColumn },
            ],
        },
    ],

    addBadPixel => [
        pixels => {
            loop => { offset => 48, nbits => 32 },
            over => [
                ccdId  	  => { value => [  0,  4 ], fmt => '%d', isa => CcdId 	  },
                ccdRow 	  => { value => [  4, 10 ], fmt => '%d', isa => CcdColumn },
                ccdColumn => { value => [ 14, 10 ], fmt => '%d', isa => CcdRow 	  },
            ],
        },

    ],

    addPatch => [
        'patchId'      => { value => [ 48, 16 ], fmt => '%d' 	 },
        'writeAddress' => { value => [ 64, 32 ], fmt => '0x%08x' },
        'writeData'    => { array => [ 96, 32 ], fmt => '0x%08x' },
    ],


    changeConfigSetting => [
        entries => {
            loop => { offset => 48, nbits => 32 },
            over => [
                itemId    => { value => [  0, 16 ], fmt => '%d' },
                itemValue => { value => [ 16, 16 ], fmt => '%d' },
            ],
        },
    ],

    continueUpload => [

        continuationData => { array => [ 64, 32 ], fmt => '0x%08x' },

    ],

    executeBep => [

        functionAddress   => { value => [ 64, 32 ], fmt => '0x%08x' },
        functionArguments => { array => [ 96, 32 ], fmt => '0x%08x' },
    ],

    executeFep => [

        fepId 		  => { value => [ 48,  3 ], fmt => '%d', isa => FepId },
        functionAddress   => { value => [ 64, 32 ], fmt => '0x%08x' 	      },
        functionArguments => { array => [ 96, 32 ], fmt => '0x%08x' 	      },
    ],

    load1dBlock => [

        windowSlotIndex => { value => [ 48, 16 ], fmt => '%d' 	  },
        checksum        => { value => [ 64, 16 ], fmt => '%d' 	  },
        windowBlockId   => { value => [ 80, 32 ], fmt => '0x%08x' },

        windows => {
            loop => { offset => 112, nbits => 60 },
            over => [
                ccdId 		    => { value => [  0,  4 ],  fmt => '%d', isa => CcdId     	},
                ccdColumn 	    => { value => [  4, 10 ],  fmt => '%d', isa => CcdColumn 	},
                width               => { value => [ 14, 10 ],  fmt => '%d', isa => ColumnCount  },
                sampleCycle         => { value => [ 24,  8 ],  fmt => '%d' 		     	},
                lowerEventAmplitude => { value => [ 32, 12 ],  fmt => '%d' 		     	},
                eventAmplitudeRange => { value => [ 44, 16 ],  fmt => '%d' 		     	},
            ],
        },

    ],

    load2dBlock => [

        windowSlotIndex => { value => [ 48, 16 ], fmt => '%d' 	  },
        checksum        => { value => [ 64, 16 ], fmt => '%d' 	  },
        windowBlockId   => { value => [ 80, 32 ], fmt => '0x%08x' },
        windows         => {
            loop => { offset => 112, nbits => 80 },
            over => [
                ccdId  		    => { value => [  0,  4 ],  fmt => '%d', isa => CcdId     	},
                ccdRow 		    => { value => [  4, 10 ],  fmt => '%d', isa => CcdRow    	},
                ccdColumn 	    => { value => [ 14, 10 ],  fmt => '%d', isa => CcdColumn 	},
                width               => { value => [ 24, 10 ],  fmt => '%d', isa => ColumnCount  },
                height              => { value => [ 34, 10 ],  fmt => '%d', isa => RowCount    	},
                sampleCycle         => { value => [ 44,  8 ],  fmt => '%d' 		     	},
                lowerEventAmplitude => { value => [ 52, 12 ],  fmt => '%d' 		     	},
                eventAmplitudeRange => { value => [ 64, 16 ],  fmt => '%d' 		     	},
            ],
        },
    ],

    loadCcBlock => [


        ccBlockSlotIndex 	=> { value => [ 48, 16 ], 	 fmt => '%d' 				   },
        checksum         	=> { value => [ 64, 16 ], 	 fmt => '%d' 				   },
        parameterBlockId 	=> { value => [ 80, 32 ], 	 fmt => '0x%08x' 			   },
        fepCcdSelect     	=> { array => [ 112, 4, 6 ], 	 fmt => '%4d', 	  isa => FepCcdSelect 	   },
        fepMode               	=> { value => [ 136, 4 ], 	 fmt => '%d' 				   },
        bepPackingMode        	=> { value => [ 140, 4 ], 	 fmt => '%d' 				   },
        ignoreBadColumnMap    	=> { value => [ 144, 1 ], 	 fmt => '%d' 				   },
        recomputeBias         	=> { value => [ 145, 1 ], 	 fmt => '%d' 				   },
        trickleBias           	=> { value => [ 146, 1 ], 	 fmt => '%d' 				   },
        rowSum                	=> { value => [ 147, 4 ], 	 fmt => '%d' 				   },
        columnSum             	=> { value => [ 151, 4 ], 	 fmt => '%d' 				   },
        overclockPairsPerNode 	=> { value => [ 155, 4 ], 	 fmt => '%d' 				   },
        outputRegisterMode    	=> { value => [ 159, 2 ], 	 fmt => '%d' 				   },
        ccdVideoResponse 	=> { array => [ 161,  1, 6 ], 	 fmt => '%d' 				   },
        fep0EventThreshold 	=> { array => [ 176, 16, 4, 1 ], fmt => '%4d', 	  isa => FepEventThreshold },
        fep1EventThreshold 	=> { array => [ 240, 16, 4, 1 ], fmt => '%4d', 	  isa => FepEventThreshold },
        fep2EventThreshold 	=> { array => [ 304, 16, 4, 1 ], fmt => '%4d', 	  isa => FepEventThreshold },
        fep3EventThreshold 	=> { array => [ 368, 16, 4, 1 ], fmt => '%4d', 	  isa => FepEventThreshold },
        fep4EventThreshold 	=> { array => [ 432, 16, 4, 1 ], fmt => '%4d', 	  isa => FepEventThreshold },
        fep5EventThreshold 	=> { array => [ 496, 16, 4, 1 ], fmt => '%4d', 	  isa => FepEventThreshold },
        fep0SplitThreshold 	=> { array => [ 560, 16, 4 ], 	 fmt => '%4d', 	  isa => FepSplitThreshold },
        fep1SplitThreshold 	=> { array => [ 624, 16, 4 ], 	 fmt => '%4d', 	  isa => FepSplitThreshold },
        fep2SplitThreshold 	=> { array => [ 688, 16, 4 ], 	 fmt => '%4d', 	  isa => FepSplitThreshold },
        fep3SplitThreshold 	=> { array => [ 752, 16, 4 ], 	 fmt => '%4d', 	  isa => FepSplitThreshold },
        fep4SplitThreshold 	=> { array => [ 816, 16, 4 ], 	 fmt => '%4d', 	  isa => FepSplitThreshold },
        fep5SplitThreshold 	=> { array => [ 880, 16, 4 ], 	 fmt => '%4d', 	  isa => FepSplitThreshold },
        lowerEventAmplitude     => { value => [ 944,  16 ], 	 fmt => '%d' 				   },
        eventAmplitudeRange     => { value => [ 960,  16 ], 	 fmt => '%d' 				   },
        gradeSelections         => { value => [ 976,   4 ],  	 fmt => '0x%04x', isa => GradeSelection    },
        windowSlotIndex         => { value => [ 992,  16 ], 	 fmt => '%d' 				   },
        rawCompressionSlotIndex => { value => [ 1008,  8 ],  	 fmt => '%d' 				   },
        ignoreInitialFrames     => { value => [ 1024, 16 ], 	 fmt => '%d' 				   },
        biasAlgorithmId 	=> { array => [ 1040,  8, 6 ],   fmt => '%4d' 				   },
        biasRejection   	=> { array => [ 1088, 16, 6 ], 	 fmt => '%4d' 				   },
        fep0VideoOffset 	=> { array => [ 1184, 16, 4 ], 	 fmt => '%4d' 				   },
        fep1VideoOffset 	=> { array => [ 1248, 16, 4 ], 	 fmt => '%4d' 				   },
        fep2VideoOffset 	=> { array => [ 1312, 16, 4 ], 	 fmt => '%4d' 				   },
        fep3VideoOffset 	=> { array => [ 1376, 16, 4 ], 	 fmt => '%4d' 				   },
        fep4VideoOffset 	=> { array => [ 1440, 16, 4 ], 	 fmt => '%4d' 				   },
        fep5VideoOffset 	=> { array => [ 1504, 16, 4 ], 	 fmt => '%4d' 				   },
        deaLoadOverride 	=> { value => [ 1568, 32 ], 	 fmt => '%d' 				   },
        fepLoadOverride 	=> { value => [ 1600, 32 ], 	 fmt => '%d' 				   },

    ],

    loadDeaBlock => [

        deaBlockSlotIndex => { value => [ 48,  16 ], fmt => '%d'     },
        checksum          => { value => [ 64,  16 ], fmt => '%d'     },
        deaBlockId        => { value => [ 80,  32 ], fmt => '0x%08x' },
        sampleRate        => { value => [ 112,  8 ], fmt => '%d'     },
        queries           => {
            loop => { offset => 128, nbits => 16 },
            over => [
                ccdId   => { value => [ 0, 8 ], fmt => '%d', isa => CcdId | Enum [10] },
                queryId => { value => [ 8, 8 ], fmt => '%d' 			      },
            ],
        },
    ],

    loadTeBlock => [

        teBlockSlotIndex 	 => { value => [  48, 16 ], 			  fmt => '%d' 				    },
        checksum         	 => { value => [  64, 16 ], 			  fmt => '%d' 				    },
        parameterBlockId 	 => { value => [  80, 32 ], 			  fmt => '0x%08x' 			    },
        fepCcdSelect 		 => { array => [ 112,  4, 6 ], 			  fmt => '%4d',    isa => FepCcdSelect 	    },
        fepMode            	 => { value => [ 136,  4 ], 			  fmt => '%d' 				    },
        bepPackingMode     	 => { value => [ 140,  4 ], 			  fmt => '%d' 				    },
        onChip2x2Summing   	 => { value => [ 144,  1 ], 			  fmt => '%d' 				    },
        ignoreBadPixelMap  	 => { value => [ 145,  1 ], 			  fmt => '%d' 				    },
        ignoreBadColumnMap 	 => { value => [ 146,  1 ], 			  fmt => '%d' 				    },
        recomputeBias      	 => { value => [ 147,  1 ], 			  fmt => '%d' 				    },
        trickleBias        	 => { value => [ 148,  1 ], 			  fmt => '%d' 				    },
        subarrayStartRow 	 => { value => [ 149, 10 ], 			  fmt => '%d', 	   isa => CcdRow 	    },
        subarrayRowCount 	 => { value => [ 159, 10 ], 			  fmt => '%d', 	   isa => RowCount 	    },
        overclockPairsPerNode 	 => { value => [ 169,  4 ], 			  fmt => '%d' 				    },
        outputRegisterMode    	 => { value => [ 173,  2 ], 			  fmt => '%d' 				    },
        ccdVideoResponse 	 => { array => [ 175,  1, 6 ], 			  fmt => '%d' 				    },

        squeegeeRows 		 => { if_true => [ 181, 3 ], value => [ 181, 3 ], fmt => '%d' 				    },
        squeegeeIdle1 		 => { if_true => [ 181, 3 ], value => [ 184, 2 ], fmt => '%d' 				    },
        squeegeeIdle2 		 => { if_true => [ 181, 3 ], value => [ 186, 2 ], fmt => '%d' 				    },
        squeegeeFlushCount 	 => { if_true => [ 181, 3 ], value => [ 188, 3 ], fmt => '%d' 				    },
        squeegeeExtraRows 	 => { if_true => [ 181, 3 ], value => [ 191, 1 ], fmt => '%d' 				    },

        primaryExposure   	 => { value => [  192, 16 ], 			  fmt => '%d' 				    },
        secondaryExposure 	 => { value => [  208, 16 ], 			  fmt => '%d' 				    },
        dutyCycle         	 => { value => [  224, 16 ], 			  fmt => '%d' 				    },
        fep0EventThreshold 	 => { array => [  240, 16, 4, 1 ], 		  fmt => '%4d',    isa => FepEventThreshold },
        fep1EventThreshold 	 => { array => [  304, 16, 4, 1 ], 		  fmt => '%4d',    isa => FepEventThreshold },
        fep2EventThreshold 	 => { array => [  368, 16, 4, 1 ], 		  fmt => '%4d',    isa => FepEventThreshold },
        fep3EventThreshold 	 => { array => [  432, 16, 4, 1 ], 		  fmt => '%4d',    isa => FepEventThreshold },
        fep4EventThreshold 	 => { array => [  496, 16, 4, 1 ], 		  fmt => '%4d',    isa => FepEventThreshold },
        fep5EventThreshold 	 => { array => [  560, 16, 4, 1 ], 		  fmt => '%4d',    isa => FepEventThreshold },
        fep0SplitThreshold 	 => { array => [  624, 16, 4 ], 		  fmt => '%4d',    isa => FepSplitThreshold },
        fep1SplitThreshold 	 => { array => [  688, 16, 4 ], 		  fmt => '%4d',    isa => FepSplitThreshold },
        fep2SplitThreshold 	 => { array => [  752, 16, 4 ], 		  fmt => '%4d',    isa => FepSplitThreshold },
        fep3SplitThreshold 	 => { array => [  816, 16, 4 ], 		  fmt => '%4d',    isa => FepSplitThreshold },
        fep4SplitThreshold 	 => { array => [  880, 16, 4 ], 		  fmt => '%4d',    isa => FepSplitThreshold },
        fep5SplitThreshold 	 => { array => [  944, 16, 4 ], 		  fmt => '%4d',    isa => FepSplitThreshold },
        lowerEventAmplitude 	 => { value => [ 1008, 16 ], 			  fmt => '%d' 				    },
        eventAmplitudeRange 	 => { value => [ 1024, 16 ], 			  fmt => '%d' 				    },
        gradeSelections 	 => { array => [ 1040, 32, 8 ], 		  fmt => '0x%08x', isa => GradeSelections   },
        windowSlotIndex 	 => { value => [ 1296, 16 ], 			  fmt => '%d' 				    },
        histogramCount  	 => { value => [ 1312, 16 ], 			  fmt => '%d' 				    },
        biasCompressionSlotIndex => { array => [ 1328,  8, 6 ], 		  fmt => '%4d' 				    },
        rawCompressionSlotIndex  => { value => [ 1376,  8 ],  			  fmt => '%d' 				    },
        ignoreInitialFrames      => { value => [ 1392, 16 ], 			  fmt => '%d' 				    },
        biasAlgorithmId 	 => { array => [ 1408,  8, 6 ], 		  fmt => '%4d' 				    },
        biasArg0 		 => { array => [ 1472, 16, 6, 1 ], 		  fmt => '%4d' 				    },
        biasArg1 		 => { array => [ 1568, 16, 6, 1 ], 		  fmt => '%4d' 				    },
        biasArg2 		 => { array => [ 1664, 16, 6, 1 ], 		  fmt => '%4d' 				    },
        biasArg3 		 => { array => [ 1760, 16, 6, 1 ], 		  fmt => '%4d' 				    },
        biasArg4 		 => { array => [ 1856, 16, 6, 1 ], 		  fmt => '%4d' 				    },
        fep0VideoOffset 	 => { array => [ 1952, 16, 4 ], 		  fmt => '%4d' 				    },
        fep1VideoOffset 	 => { array => [ 2016, 16, 4 ], 		  fmt => '%4d' 				    },
        fep2VideoOffset 	 => { array => [ 2080, 16, 4 ], 		  fmt => '%4d' 				    },
        fep3VideoOffset 	 => { array => [ 2144, 16, 4 ], 		  fmt => '%4d' 				    },
        fep4VideoOffset 	 => { array => [ 2208, 16, 4 ], 		  fmt => '%4d' 				    },
        fep5VideoOffset 	 => { array => [ 2272, 16, 4 ], 		  fmt => '%4d' 				    },
        deaLoadOverride 	 => { value => [ 2336, 32 ], 			  fmt => '0x%08x' 			    },
        fepLoadOverride 	 => { value => [ 2368, 32 ], 			  fmt => '0x%08x' 			    },
    ],

    readBep => [
        readAddress 	  => { value => [ 48, 32 ],  fmt => '0x%08x' 		},
        wordCount   	  => { value => [ 80, 32 ],  fmt => '%d'     		},
    ],

    readFep => [
        fepId       	  => { value => [ 48, 3 ],   fmt => '%d', isa => FepId 	},
        readAddress 	  => { value => [ 64, 32 ],  fmt => '0x%08x' 		},
        wordCount   	  => { value => [ 96, 32 ],  fmt => '%d' 		},
    ],

    readPram => [
        ccdId     	  => { value => [ 48,  4 ],  fmt => '%d', isa => CcdId  },
        readIndex 	  => { value => [ 64, 16 ],  fmt => '0x%04x' 	       	},
        wordCount 	  => { value => [ 80, 16 ],  fmt => '%d' 	       	},
    ],

    readSram => [
        ccdId     	  => { value => [ 48,  4 ],  fmt => '%d', isa => CcdId  },
        readIndex 	  => { value => [ 64, 16 ],  fmt => '0x%04x' 	       	},
        wordCount 	  => { value => [ 80, 16 ],  fmt => '%d' 	       	},
    ],

    removePatches => [
        patchIds 	  => { array => [ 48, 16 ],  fmt => '%d' 		},
    ],


    startScience => [
        blockSlotIndex 	  => { value => [ 48, 16 ],  fmt => '%d' 		},
    ],

    startDea => [
        deaBlockSlotIndex => { value => [ 48, 16 ],  fmt => '%d' 		},
    ],

    startUpload => [
        loadAddress    	  => { value => [ 64,  32 ], fmt => '0x%08x' 		},
        totalCount     	  => { value => [ 96,  32 ], fmt => '%d' 	  	},
        executeAddress 	  => { value => [ 128, 32 ], fmt => '0x%08x' 		},
        initialData    	  => { array => [ 160, 32 ], fmt => '0x%08x' 		},
    ],

    writeBep => [
        writeAddress 	  => { value => [ 64, 32 ],  fmt => '0x%08x' 		},
        writeData    	  => { array => [ 96, 32 ],  fmt => '0x%08x' 		},
    ],

    writeFep => [
        fepId        	  => { value => [ 48,  3 ],  fmt => '%d', isa => FepId 	},
        writeAddress 	  => { value => [ 64, 32 ],  fmt => '0x%08x' 	  	},
        writeData    	  => { array => [ 96, 32 ],  fmt => '0x%08x' 	  	},
    ],

    writePram => [
        ccdId      	  => { value => [ 48,  4 ],  fmt => '%d', isa => CcdId 	},
        writeIndex 	  => { value => [ 64, 16 ],  fmt => '0x%04x' 		},
        writeData  	  => { array => [ 80, 16 ],  fmt => '0x%04x' 		},
    ],

    writeSram => [
        ccdId      	  => { value => [ 48,  4 ],  fmt => '%d', isa => CcdId 	},
        writeIndex 	  => { value => [ 64, 16 ],  fmt => '0x%04x' 	       	},
        writeData  	  => { array => [ 80, 16 ],  fmt => '0x%04x' 	       	},
    ],

    fifoFlush => [
	padding    	  => { array => [ 48, 16 ],  fmt => '%d' 		}
     ],

);

#>>> no perltidy

# input: starting offset in bits, length
#
#  Algorithm derived from Royce Buehler's lcmd bit routine.  Bits are in original
#  order as given in command stream.
#
#  1. Find 16 bit word containing the offset, index = offset >> 4.
#
#  2. The LSB bit requested is bit = offset & 15, where bit = 0 is the
#     *rightmost* bit in the word
#
#  3. If more bits are required than are available to the left of the
#     LSB bit in the current word, take them from the *next* word,
#     starting at the rightmost bit.

sub bits {

    #  my ( $vec, $start, $len ) = @_;

    my ( $start, $len ) = splice( @_, 1 );

    croak( "starting bit ($start) too large\n" )
      if $start >= length( $_[0]) * 16;

    croak( "length($len) too large\n" )
      if $start + $len > length( $_[0]) * 16;

    my $value = 0;

    my $iword = $start >> 4;

    my $bit = $start & 0x000f;

    my $shift = 0;
    for ( ; $len ; $iword++, $bit = 0) {

	my $word = vec( $_[0], $iword, 16 );

	# shift off all bits to the right of the requested LSB
	$word >>= $bit;

	# need more bits than are available to the left of the
	# requested LSB in this word?
	if ( $bit + $len > 16  ) {

	    # no need to mask as the bits are all significant
	    $value |= $word << $shift;

	    my $bits_used = 16 - $bit;

	    $len -= $bits_used;

	    $shift += $bits_used;
	}

	# enough bits to complete request
	else {

	    # mask off unwanted higher order bits and move
	    # left to accomodate any previous bits added
	    $value |= ( $word &= 0xffff >> ( 16 - $len ) ) << $shift;

	    last;

	}

    }

    return $value;
}

sub loadBlock {

    my ( $buf ) = @_;

    my $block = _extract( $buf, $decode{commandHeader}, 0, 0 );

    my $cmd = $CMD{ $block->{commandOpcode} }
      or croak( "illegal command op: $block->{commandOpCode}\n" );

    my $bit_length = $block->{commandLength} * 16;

    # not all commands take extra parameters; those that do
    # will have entries in %decode.
    if ( $decode{ $cmd->{name} } ) {

        my $params;

        try {

            $params = _extract( $buf, $decode{ $cmd->{name} }, $bit_length, 0 );
        }
        catch {

            croak( "$cmd->{name}: $_\n" );

        };

	@{$block}{ keys $params } = values $params;

    }

    $block->{commandName} = $cmd->{name};
    return $block;
}

sub _extract {

    # my ( $buf, $spec, $block_length, $block_offset ) = @_ ;

    my ( $block_spec, $block_length, $block_offset ) = splice( @_, 1 );

    my @block_spec = @$block_spec;

    my %block;

    while ( @block_spec ) {

        my ( $attr, $spec ) = splice( @block_spec, 0, 2 );

	my $value;

        if ( $spec->{loop} ) {

            my $loop = $spec->{loop};
            my $over = $spec->{over}
              or croak( "internal error: $attr: loop without over\n" );

            croak( "internal error: $attr: loop nbits not specified\n" )
              unless $loop->{nbits};

            my $n = ( $block_length - $loop->{offset} ) / $loop->{nbits};

            my @values;

            for my $idx ( 0 .. $n - 1 ) {

                #                printf( "%-28.28s[%d]\n", $attr, $idx );

                push @values,
                  _extract( $_[0], $over, $loop->{nbits},
                    $block_offset + $loop->{offset} + $loop->{nbits} * $idx );

            }
	    $block{$attr} = \@values;

            next;
        }

        elsif ( $spec->{if_true} ) {

            my ( $offset, $nbits ) = @{ $spec->{if_true} };

            next unless bits( $_[0], $block_offset + $offset, $nbits );

        }

        my $fmt = $spec->{fmt};

        if ( $spec->{value} ) {

            my ( $offset, $nbits ) = @{ $spec->{value} };

            $value = bits( $_[0], $block_offset + $offset, $nbits );

            #            printf( "%-28.28s = $fmt\n", $attr, $value );

        }

        elsif ( $spec->{array} ) {

            my ( $offset, $nbits, $n, $signed ) = @{ $spec->{array} };

            # if the value should be signed, create the
            # operand which will make it thus
            my $neg = $signed ? 1 << ( $nbits - 1 ) : 0;

            $n //= ( $block_length - $offset ) / $nbits;

            my @values;

            for my $idx ( 0 .. $n - 1 ) {

                my $val = bits( $_[0], $offset + $idx * $nbits, $nbits );

                $val -= 2 * $neg if $neg && $val >= $neg;

                push @values, $val;

            }

	    $value = \@values;
 #            printf( "%-28.28s = " . ( "$fmt " x $n ) . "\n", $attr, @values );

        }

        else {

            croak( "$attr: can't figure out attribute type\n" );
        }

        try {

            validate( [ $value ], $spec->{isa} )
              if $spec->{isa};

        }
        catch {

            croak( "$attr: illegal value ($value): $_\n" );

        };

	$block{$attr} = $value;

    }

    return \%block;
}


1;


=head1 NAME

CXC::ACIS::Command::Block - decode raw ACIS Command blocks

=head1 SYNOPSIS

  use CXC::ACIS::Command::Block qw[ loadBlock ];

  my $block = loadBlock( $raw_block );

=head1 DESCRIPTION

This module provides the B<loadBlock> function, which decodes a raw
ACIS command block as read by
B<L<CXC::ACIS::Command::OFLS::loadOFLS_dat>>.

=head1 FUNCTIONS

=head2 loadBlock

  my $block = loadBlock( $raw_block );

This returns a hashref containing the parsed command block.  See
L</RECOGNIZED BLOCKS AND THEIR CONTENTS> for the returned fields.


=head1 RECOGNIZED BLOCKS AND THEIR CONTENTS

=head2 Common fields

Each block contains the following fields:

=over

=item *

C<commandLength> I<integer>

=item *

C<commandIdentifier> I<integer>

=item *

C<commandOpcode>  I<integer>

=item *

C<commandName>  I<string>

The name of the command.

=back

The commands which are recognized (and the values of the C<commandName> field are:

=head2 addBadColumn

=over

=item *

columns I<variable length array of hashes>

=over

=item *

C<ccdId>  I<integer>

=item *

C<ccdColumn> I<integer>

=back

=back

=head2 addBadPixel

=over

=item * pixels I<variable length array of hashes>

=over

=item *

C<ccdId> I<integer>

=item *

C<ccdRow> I<integer>

=item *

C<ccdColumn> I<integer>

=back

=back

=head2 addPatch

=over

=item *

C<patchId> I<integer>

=item *

C<writeAddress> I<integer>

=item *

C<writeData> I<variable length array of integers>


=back

=head2 changeConfigSetting


=over

=item *

C<entries> I<variable length array of hashes>

=over

=item *
C<itemId> I<integer>

=item *
C<itemValue> I<integer>

=back

=back

=head2 continueUpload

=over

=item *

C<continuationData> I<variable length array of integers>

=back

=head2 executeBep

=over

=item *

C<functionAddress>  I<integer>

=item *

C<functionArguments> I<variable length array of integers>

=back

=head2 executeFep

=over

=item *

C<fepId> I<integer>


=item *

C<functionAddress> I<integer>

=item *

C<functionArguments> I<variable length array of integers>

=back

=head2 load1dBlock

=over

=item *

C<windowSlotIndex> I<integer>

=item *

C<checksum> I<integer>


=item *

C<windowBlockId> I<integer>


=item *

C<windows> I<variable length array of hashes>

=over

=item *
C<ccdId> I<integer>

=item *
C<ccdColumn> I<integer>

=item *
C<width> I<integer>

=item *
C<sampleCycle> I<integer>

=item *
C<lowerEventAmplitude> I<integer>

=item *
C<eventAmplitudeRange> I<integer>


=back

=back


=head2 load2dBlock

=over

=item *

C<windowSlotIndex> I<integer>


=item *

C<checksum> I<integer>


=item *

C<windowBlockId> I<integer>

=item *

C<windows> I<variable length array of hashes>

=over

=item *
C<ccdId> I<integer>

=item *
C<ccdRow> I<integer>

=item *
C<ccdColumn> I<integer>

=item *
C<width> I<integer>

=item *
C<height> I<integer>

=item *
C<sampleCycle> I<integer>

=item *
C<lowerEventAmplitude> I<integer>

=item *
C<eventAmplitudeRange> I<integer>


=back

=back

=head2 loadCcBlock

=over

=item *

C<cBlockSlotIndex> I<integer>


=item *

C<checksum> I<integer>


=item *

C<parameterBlockId> I<integer>


=item *

C<fepCcdSelect> I<fixed length (6) array of integers>

=item *

C<fepMode> I<integer>


=item *

C<bepPackingMode> I<integer>


=item *

C<ignoreBadColumnMap> I<integer>

=item *

C<recomputeBias> I<integer>

=item *

C<trickleBias> I<integer>

=item *

C<rowSum> I<integer>

=item *

C<columnSum> I<integer>

=item *

C<overclockPairsPerNode> I<integer>

=item *

C<outputRegisterMode> I<integer>

=item *

C<ccdVideoResponse> I<fixed length (6) array of integers>

=item *

C<fep0EventThreshold> I<fixed length (4) array of integers>

=item *

C<fep1EventThreshold> I<fixed length (4) array of integers>

=item *

C<fep2EventThreshold> I<fixed length (4) array of integers>

=item *

C<fep3EventThreshold> I<fixed length (4) array of integers>

=item *

C<fep4EventThreshold> I<fixed length (4) array of integers>

=item *

C<fep5EventThreshold> I<fixed length (4) array of integers>

=item *

C<fep0SplitThreshold> I<fixed length (4) array of integers>

=item *

C<fep1SplitThreshold> I<fixed length (4) array of integers>

=item *

C<fep2SplitThreshold> I<fixed length (4) array of integers>

=item *

C<fep3SplitThreshold> I<fixed length (4) array of integers>

=item *

C<fep4SplitThreshold> I<fixed length (4) array of integers>

=item *

C<fep5SplitThreshold> I<fixed length (4) array of integers>


=item *

C<lowerEventAmplitude> I<integer>

=item *

C<eventAmplitudeRange> I<integer>

=item *

C<gradeSelections> I<integer>

=item *

C<windowSlotIndex> I<integer>

=item *

C<rawCompressionSlotIndex> I<integer>

=item *

C<ignoreInitialFrames> I<integer>

=item *

C<biasAlgorithmId> I<fixed length (6) array of integers>

=item *

C<biasRejection> I<fixed length (6) array of integers>

=item *

C<fep0VideoOffset> I<fixed length (4) array of integers>

=item *

C<fep1VideoOffset> I<fixed length (4) array of integers>

=item *

C<fep2VideoOffset> I<fixed length (4) array of integers>

=item *

C<fep3VideoOffset> I<fixed length (4) array of integers>

=item *

C<fep4VideoOffset> I<fixed length (4) array of integers>

=item *

C<fep5VideoOffset> I<fixed length (4) array of integers>

=item *

C<deaLoadOverride> I<integer>

=item *

C<fepLoadOverride> I<integer>

=back

=head2 loadDeaBlock

=over

=item *

C<deaBlockSlotIndex> I<integer>

=item *

C<checksum> I<integer>

=item *

C<deaBlockId> I<integer>

=item *

C<sampleRate> I<integer>

=item *

C<queries> I<variable length array of hashes>

=over

=item *
C<ccdId> I<integer>

=item *
C<queryId> I<integer>

=back

=back

=head2 loadTeBlock

=over

=item *

C<teBlockSlotIndex> I<integer>

=item *

C<checksum> I<integer>

=item *

C<parameterBlockId> I<integer>

=item *

C<fepCcdSelect> I<fixed length (6) array of integers>

=item *

C<fepMode> I<integer>

=item *

C<bepPackingMode> I<integer>

=item *

C<onChip2x2Summing> I<integer>

=item *

C<ignoreBadPixelMap> I<integer>

=item *

C<ignoreBadColumnMap> I<integer>

=item *

C<recomputeBias> I<integer>

=item *

C<trickleBias> I<integer>

=item *

C<subarrayStartRow> I<integer>

=item *

C<subarrayRowCount> I<integer>

=item *

C<overclockPairsPerNode> I<integer>

=item *

C<outputRegisterMode> I<integer>

=item *

C<ccdVideoResponse> I<fixed length (6) array of integers>

=item *

C<squeegeeIdle1> I<integer> I<[optional]>

=item *
C<squeegeeIdle2> I<integer> I<[optional]>

=item *
C<squeegeeFlushCount> I<integer> I<[optional]>

=item *
C<squeegeeExtraRows> I<integer> I<[optional]>

=item *

C<primaryExposure> I<integer>

=item *

C<secondaryExposure> I<integer>

=item *

C<dutyCycle> I<integer>

=item *

C<fep0EventThreshold> I<fixed length (4) array of integers>

=item *

C<fep1EventThreshold> I<fixed length (4) array of integers>

=item *

C<fep2EventThreshold> I<fixed length (4) array of integers>

=item *

C<fep3EventThreshold> I<fixed length (4) array of integers>

=item *

C<fep4EventThreshold> I<fixed length (4) array of integers>

=item *

C<fep5EventThreshold> I<fixed length (4) array of integers>

=item *

C<fep0SplitThreshold> I<fixed length (4) array of integers>

=item *

C<fep1SplitThreshold> I<fixed length (4) array of integers>

=item *

C<fep2SplitThreshold> I<fixed length (4) array of integers>

=item *

C<fep3SplitThreshold> I<fixed length (4) array of integers>

=item *

C<fep4SplitThreshold> I<fixed length (4) array of integers>

=item *

C<fep5SplitThreshold> I<fixed length (4) array of integers>

=item *

C<lowerEventAmplitude> I<integer>

=item *

C<eventAmplitudeRange> I<integer>

=item *

C<gradeSelections> I<fixed length (8) array of integers>


=item *

C<windowSlotIndex> I<integer>

=item *

C<histogramCount> I<integer>

=item *

C<biasCompressionSlotIndex> I<fixed length (6) array of integers>


=item *

C<rawCompressionSlotIndex> I<integer>

=item *

C<ignoreInitialFrames> I<integer>

=item *

C<biasAlgorithmId> I<fixed length (6) array of integers>

=item *

C<biasArg0> I<fixed length (6) array of integers>

=item *

C<biasArg1> I<fixed length (6) array of integers>

=item *

C<biasArg2> I<fixed length (6) array of integers>

=item *

C<biasArg3> I<fixed length (6) array of integers>

=item *

C<biasArg4> I<fixed length (6) array of integers>

=item *

C<fep0VideoOffset> I<fixed length (4) array of integers>

=item *

C<fep1VideoOffset> I<fixed length (4) array of integers>

=item *

C<fep2VideoOffset> I<fixed length (4) rray of integers>

=item *

C<fep3VideoOffset> I<fixed length (4) array of integers>

=item *

C<fep4VideoOffset> I<fixed length (4) array of integers>

=item *

C<fep5VideoOffset> I<fixed length (4) array of integers>

=item *

C<deaLoadOverride> I<integer>

=item *

C<fepLoadOverride> I<integer>

=back

=head2 readBep

=over


=item *

C<readAddress> I<integer>

=item *

C<wordCount> I<integer>

=back

=head2 readFep

=over

=item *

C<fepId> I<integer>

=item *

C<readAddress> I<integer>

=item *

C<wordCount> I<integer>

=back

=head2 readPram

=over

=item *

C<ccdId> I<integer>

=item *

C<readIndex> I<integer>

=item *

C<wordCount> I<integer>

=back

=head2 readSram

=over


=item *

C<ccdId> I<integer>

=item *

C<readIndex> I<integer>

=item *

C<wordCount> I<integer>

=back

=head2 removePatches

=over


=item *

C<patchIds> I<variable length array of integers>


=back


=head2 startScience

=over


=item *

C<blockSlotIndex> I<integer>

=back

=head2 startDea

=over


=item *

C<deaBlockSlotIndex> I<integer>

=back

=head2 startUpload

=over


=item *

C<loadAddress> I<integer>

=item *

C<totalCount> I<integer>

=item *

C<executeAddress> I<integer>


=item *

C<initialData> I<variable length array of integers>


=back

=head2 writeBep

=over

=item *

C<writeAddress> I<integer>


=item *

C<writeData> I<variable length array of integers>


=back

=head2 writeFep

=over

=item *

C<fepId> I<integer>

=item *

C<writeAddress> I<integer>


=item *

C<writeData> I<variable length array of integers>

=back

=head2 writePram

=over

=item *

C<ccdId> I<integer>

=item *

C<writeIndex> I<integer>


=item *

C<writeData> I<variable length array of integers>


=back

=head2 writeSram

=over

=item *

C<ccdId> I<integer>

=item *

C<writeIndex> I<integer>


=item *

C<writeData> I<variable length array of integers>


=back

=head2 fifoFlush

=over

=item *

C<padding> I<variable length array of integers>

=back

=head1 LICENSE AND COPYRIGHT

Copyright (c) 2014 The Smithsonian Astrophysical Observatory

B<CXC::ACIS::Command::Block> is free software: you can redistribute it
and/or modify it under the terms of the GNU General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

=head1 AUTHOR

Diab Jerius  E<lt>djerius@cfa.harvard.eduE<gt>
