package CXC::ACIS::Command::Dump;

use 5.10.1;

use strict;
use warnings;

our $VERSION = '0.02';

use Carp;

use Pegex::Parser;
use CXC::ACIS::Command::Dump::Grammar;
use CXC::ACIS::Command::Dump::Tree;
use Exporter qw[ import ];

use Types::Standard qw[ Enum ];
use Type::Params qw[ compile ];

our @EXPORT_OK = qw[ loadMIT loadHTML loadDump ];

## no critic (ProhibitAccessOfPrivateData)

sub loadMIT {

    my $tag = shift;

    require HTTP::Tiny;

    my $url = "http://acis.mit.edu/cgi-bin/get-atbls?tag=$tag";

    my $response = HTTP::Tiny->new->get( $url );
    croak(
        sprintf(
            "error retrieving tag $tag: status( %s ), %s\n",
            $response->{status}, $response->{reason} )
    ) unless $response->{success};


    return loadHTML( $response->{content} );
}

sub loadHTML {

    require HTML::TableParser;

    # my $html = shift;

    my $parser = Pegex::Parser->new(
        grammar  => CXC::ACIS::Command::Dump::Grammar->new,
        receiver => CXC::ACIS::Command::Dump::Tree->new,
    );

    my ( @hdr, %hdr );

    my $meta;
    my @commands;

    # the html table has the format
    #  ( header row row )+
    # the first row has metadata; the second has the actual Command::Dump data

    my $hdr = sub {

        my ( $tbl_id, $line_no, $data, $udata ) = @_;

        @hdr  = @$data;
        $meta = 1;

    };

    my $row = sub {

        my ( $tbl_id, $line_no, $data, $udata ) = @_;

        if ( $meta ) {

            @hdr{@hdr} = @$data;

            $meta = 0;

        }

        else {

            my $blocks = $parser->parse( $data->[0] );

	    croak( "incorrect number of blocks; got @{[ scalar @$blocks ]}, expected 1\n" )
	      if @$blocks != 1;

            push @commands, _cleanup( \%hdr, $blocks->[0] );
        }


    };


    HTML::TableParser->new( [ {
                id  => 'DEFAULT',
                hdr => $hdr,
                row => $row,
            } ]
      )->parse( $_[0] )
      or croak( "error parsing HTML\n" );

    return \@commands;

}

sub loadDump {

    my $parser = Pegex::Parser->new(
        grammar  => CXC::ACIS::Command::Dump::Grammar->new,
        receiver => CXC::ACIS::Command::Dump::Tree->new,
    );

    return $parser->parse( $_[0] );
}


sub _cleanup {

    return $_[1]


}

1;

__END__

=head1 NAME

CXC::ACIS::Command::Dump - parse dumps of ACIS Command blocks and sequences of blocks


=head1 SYNOPSIS

  use CXC::ACIS::Command::Dump qw[ loadMIT loadHTML loadDump ];

  # grab command dump from acis.mit.edu
  my $blocks = loadMIT( $tag ) ; # $tag is an SI mode or Block ID

  # parse HTML (downloaded from acis.mit.edu)
  my $blocks = loadHTML( $html ) ;

  # parse output of sacgs lcmd
  my $blocks = loadDump( $sacgs_lcmd_output );


=head1 DESCRIPTION

This module provides several means of parsing "dumps" of one or more
ACIS Command blocks. 


=head1 FUNCTIONS

=head2 loadMIT

  my $blocks = loadMIT( $tag );

Scrape and parse L<http://acis.mit.edu/cgi-bin/get-atbls>, passing the
C<$tag> parameter as the I<tag> query parameter.  C<$tag> may be any
tag accepted by the web site.

The parsed blocks are returned as a reference to a list of blocks,
encoded as hashes (see L<B<CXC::ACIS::Command::Block>> for their
structure).

=head2 loadHTML

  my $blocks = loadHTML( $html );

Parse an HTML representation (essentially the HTML produced by
L<http://acis.mit.edu/cgi-bin/get-atbls>) of one or more ACIS Command
Blocks.

The parsed blocks are returned as a reference to a list of blocks,
encoded as hashes (see L<B<CXC::ACIS::Command::Block>> for their
structure).

=head2 loadDump

  my $blocks = loadDump( $html );

Parse a text representation (generated by the C<sacgs/bin/lcmd> program)
of one or more ACIS Command Blocks.

The parsed blocks are returned as a reference to a list of blocks,
encoded as hashes (see L<B<CXC::ACIS::Command::Block>> for their
structure).

=head1 LICENSE AND COPYRIGHT

Copyright (c) 2014 The Smithsonian Astrophysical Observatory

B<CXC::ACIS::Command::Dump> is free software: you can redistribute it
and/or modify it under the terms of the GNU General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

=head1 AUTHOR

Diab Jerius  E<lt>djerius@cfa.harvard.eduE<gt>
