package CXC::ACIS::Command::FITS;

use strict;
use warnings;

our $VERSION = '0.02';

use Carp;

use Astro::FITS::CFITSIO::Simple qw[ rdfits ];
use PDL::Lite;

use Exporter qw[ import ];

our @EXPORT_OK = qw[ loadFITS ];

## no critic (ProhibitAccessOfPrivateData)

my %ParamBlock = (
    TE => {

        keyword => {

            # teBlockSlotIndex
            # checksum

            # squeegeeRows
            # squeegeeIdle1
            # squeegeeIdle2
            # squeegeeFlushCount
            # squeegeeExtraRows

            # WIND_ID	windowSlotIndex

            qw[
              PBLOCK	parameterBlockId
              FEP_MODE	fepMode
              BEP_MODE	bepPackingMode
              SUM_2X2	onChip2x2Summing
              NOBADPIX	ignoreBadPixelMap
              NOBADCOL	ignoreBadColumnMap
              BIAS_CAL	recomputeBias
              SENDBIAS	trickleBias
              STARTROW	subarrayStartRow
              ROWCNT	subarrayRowCount
              OCLKPAIR	overclockPairsPerNode
              ORC_MODE	outputRegisterMode


              EXPTIMEA	primaryExposure
              EXPTIMEB	secondaryExposure
              DTYCYCLE	dutyCycle
              PHAMIN	lowerEventAmplitude
              PHARANGE	eventAmplitudeRange
              HISTCNT	histogramCount
              RCMPSLOT	rawCompressionSlotIndex
              EXP_SKIP	ignoreInitialFrames
              DEALOAD	deaLoadOverride
              FEPLOAD	fepLoadOverride
              ],
        },

        # one element per fep
        column_all_feps => { qw[
              CCD_ID    fepCcdSelect
              BCMPSLOT  biasCompressionSlotIndex
              VIDRESP   ccdVideoResponse
              BIASALG	biasAlgorithmId
              BIASARG0	biasArg0
              BIASARG1	biasArg1
              BIASARG2	biasArg2
              BIASARG3	biasArg3
              BIASARG4	biasArg4
              ],
        },

        # one column per fep
        column_per_fep => { qw[
              EVT_THR   fep%dEventThreshold
              SPL_THR   fep%dSplitThreshold
              VID_OFF   fep%dVideoOffset
              ] }
    },

    CC => {
        #      ccBlockSlotIndex
        #      checksum
        #      WIND_ID	windowSlotIndex
        keyword => { qw[
              PBLOCK	parameterBlockId
              FEP_MODE	fepMode
              BEP_MODE	bepPackingMode
              NOBADCOL	ignoreBadColumnMap
              BIAS_CAL	recomputeBias
              SENDBIAS	trickleBias
              SUMROW	rowSum
              SUMCOL	columnSum
              OCLKPAIR	overclockPairsPerNode
              ORC_MODE	outputRegisterMode
              PHAMIN	lowerEventAmplitude
              PHARANGE	eventAmplitudeRange
              GRADEMAP	gradeSelections
              RCMPSLOT	rawCompressionSlotIndex
              EXP_SKIP	ignoreInitialFrames
              DEALOAD	deaLoadOverride
              FEPLOAD	fepLoadOverride
              ]
        },

        # one element per fep
        column_all_feps => { qw[
              CCD_ID    fepCcdSelect
              VIDRESP   ccdVideoResponse
              BIASALG	biasAlgorithmId
              BIASREJ	biasRejection
              ],
        },


        # one column per fep
        column_per_fep => { qw[
              EVT_THR   fep%dEventThreshold
              SPL_THR   fep%dSplitThreshold
              VID_OFF   fep%dVideoOffset
              ] }

    },

);

sub loadFITS {

    my ( $file ) = @_;

    my %data = rdfits(
        $file,
        {
            rethash => 1,
            rethdr  => 1,
        } );


    my $hdr = delete $data{_hdr};

    my $hdu = $hdr->itembyname( 'HDUCLAS2' )->value;

    if ( $hdu eq 'PARAMBLOCK' ) {

        return loadParamBlock( $hdr, \%data );

    }

    elsif ( $hdu eq 'WINDOWBLOCK' ) {
        return loadWindowBlock( $hdr, \%data );

    }

    else {

        croak( "cannot parse FITS HDU w/ HDUCLAS2=$hdu\n" )

    }

}

sub loadParamBlock {

    my ( $hdr, $columns ) = @_;

    my $naxis = $hdr->itembyname( 'NAXIS2' )->value;

    my $read_mode = $hdr->itembyname( 'READMODE' )->value;
    croak( "cannot handle READMODE=$read_mode\n" )
      unless $read_mode =~ /^(TIMED|CONTINUOUS)$/;

    $read_mode = { TIMED => 'TE', CONTINUOUS => 'CC' }->{$read_mode};
    my $map = $ParamBlock{$read_mode};

    my %block;

    if ( $read_mode eq 'TE' ) {
        $block{commandName}   = 'loadTeBlock';
        $block{commandOpcode} = 9;
    }
    elsif ( $read_mode eq 'CC' ) {
        $block{commandName}   = 'loadCcBlock';
        $block{commandOpcode} = 10;
    }

    # handle recognized HEADER keywords
    loadKeywords( $hdr, $map, \%block );


    # columns are stored per-FEP.  recreate original ACIS command block
    # structures. These all have one slot per FEP

    # initialize to 10 (no CCD)
    $block{fepCcdSelect} = [ ( 10 ) x 6 ];

    # get active FEPs
    my @fep = $columns->{'fep_id'}->list;

    for my $fitscol ( keys $map->{column_all_feps} ) {

        my $col = $map->{column_all_feps}{$fitscol};
        $block{$col} //= [ ( 0 ) x 6 ];
        $block{$col}[ $fep[$_] ] = $columns->{ lc $fitscol }->at( $_ )
          for 0 .. $#fep;

    }

    for my $fitscol ( keys $map->{column_per_fep} ) {

        my $col = $columns->{ lc $fitscol };
        my ( $nelem, $nrows ) = $col->dims;

        my $colfmt = $map->{column_per_fep}{$fitscol};

        $block{ sprintf( $colfmt, $fep[$_] ) }
          = [ $col->mslice( [], [$_] )->list ]
          for 0 .. $#fep;

        # set up default if fep is not used
        $block{ sprintf( $colfmt, $_ ) } //= [ ( 0 ) x $nelem ] for 0 .. 5;

    }


    if ( $read_mode eq 'TE' ) {

        $block{gradeSelections}
          = [ map { hex( $hdr->itembyname( "GRADEMA$_" )->value ) } 1 .. 8 ];

    }

    else {

        $block{gradeSelections} = hex( $hdr->itembyname( "GRADEMAP" )->value );
    }

    return \%block;
}

my %WindowBlock = (

    keyword => { qw[
          WIND_ID windowBlockId
          ]
    },

    gather => {
        windows => { qw[
              CCD_ID   ccdId
              LL_CCDX  ccdRow
              LL_CCDY  ccdColumn
              CCDCOL   width
              CCDROW   height
              SAMP_CYC sampleCycle
              PHAMIN   lowerEventAmplitude
              PHARANGE eventAmplitudeRange
              ],
        },
    },
 );


sub loadWindowBlock {

    my ( $hdr, $columns ) = @_;

    my %block = ( commandName => 'load2dBlock',
		  commandOpcode     => 11,
		  );

    loadKeywords( $hdr, \%WindowBlock, \%block );
    my $nrows = $hdr->itembyname( 'NAXIS2' )->value;

    gatherColumns( $nrows, $columns, $WindowBlock{gather}, \%block );

    return \%block;
}

# handle recognized HEADER keywords
sub loadKeywords {

    my ( $hdr, $map, $block ) = @_;

    for my $item ( $hdr->allitems ) {

        my $keyword = $map->{keyword}{ uc $item->keyword }
          or next;

        $block->{$keyword} = $item->value;
    }

}

sub gatherColumns {

    my ( $nrows, $columns, $map, $block ) = @_;

    for my $name ( keys %$map ) {
	$block->{$name} = [];

	for my $row ( 0..$nrows-1 ) {

	    my %record;

	    my $cmap = $map->{$name};
	    for my $col ( keys %$cmap ) {

		$record{$cmap->{$col}} = $columns->{lc $col}->at( $row );

	    }
	    push @{$block->{$name}}, \%record;
	}
    }

}

1;
__END__

=head1 NAME

CXC::ACIS::Command::FITS - parse Parameter and Window Blocks from FITS files


=head1 SYNOPSIS

  use CXC::ACIS::Command::FITS qw[ loadFITS ];

  # parse CXC DS generated parameter/window block file
  my $blocks = loadFITS( $FITS_filename );


=head1 DESCRIPTION

This module extracts parameter of window blocks from CXC DS pipeline
generated FITS files (files whose names contain C<pbk> or C<win>).


=head1 FUNCTIONS

=head2 loadFITS

  my $block = loadFITS( $FITS_filename );


Load a block from the passed FITS file.  CC and TE mode parameter
blocks (corresponding to the ACIS C<loadTeBlock> and C<loadCcBlock> commands)
and window blocks (corresponding to the ACIS C<load2dBlock> command) are recognized.
The block is returned as a hashref (see L<B<CXC::ACIS::Command::Block>> for its contents).

The following fields, are I<not> returned, as they are not stored in the FITS file:

  checksum
  commandIdentifier
  commandLength
  teBlockSlotIndex
  ccBlockSlotIndex
  windowSlotIndex


=head1 LICENSE AND COPYRIGHT

Copyright (c) 2014 The Smithsonian Astrophysical Observatory

B<CXC::ACIS::Command::FITS> is free software: you can redistribute it
and/or modify it under the terms of the GNU General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

=head1 AUTHOR

Diab Jerius  E<lt>djerius@cfa.harvard.eduE<gt>
