package CXC::ACIS::Command::OFLS;

use strict;
use warnings;

our $VERSION = '0.02';

use Exporter qw[ import ];
our @EXPORT_OK = qw[ loadOFLS_cfg loadOFLS_dat ];

use Carp;
use IO::File;

sub loadOFLS_dat {

    my $file = shift;

    my $fh = IO::File->new( $file, 'r' )
      or croak( "$file: unable to access\n" );

    $fh->binmode;


    # get file format
    $fh->seek( 15, 0 )
      or croak( "$file: truncated header\n" );

    $fh->read( my $fmt, 1 ) == 1 or croak( "error retrieving format\n" );

    my $hdrlen
      = $fmt == 2 ? 20
      : $fmt == 0 ? 16
      :             croak( "$file: unrecognized format: $fmt\n" );


    $fh->seek( 0, 0 );

    # now read the records

    my %tbl;
    my $recn;
    my $header;
    for ( $recn = 1 ; my $nr = $fh->read( $header, $hdrlen ) ; $recn++ ) {

        croak( "$file: truncated header in record $recn\n" )
          if $nr != $hdrlen;

        my ( $reclen, $id ) = unpack( "A3A10", $header );
        $reclen *= 2;

        my $body;
        croak( "$file: truncated binary in record $recn ($id)\n" )
          unless $fh->read( $body, $reclen ) == $reclen;

        my ( $ireclen ) = unpack( 'S>', $body );
        $ireclen *= 2;

        # Consistency check: internal length in body equals length from header.
        croak(
            "$file: $recn: $id: internal block length ($ireclen) doesn't match header record length ($reclen)\n"
        ) if $ireclen != $reclen;


        $tbl{$id} = $body;
    }

    return \%tbl;
}

sub loadOFLS_cfg {

    my $file = shift;

    my $fh = IO::File->new( $file, 'r' )
      or croak( "$file: unable to access\n" );

    $fh->binmode;

    my $len = ( $fh->stat )[7];

    my $cfg;
    croak( "$file: unexpected eof\n" )
      unless $len == $fh->read( $cfg, $len );

    my %tbl;
    my $recn;
    for my $record ( unpack( "(a236)*", $cfg ) ) {

	$recn++;

        my ( $id, $onblocks, @data ) = unpack( "A10 A2 (A10)16 (A4)16", $record );

	# some records have a funky number of blocks. verify
	my $nblocks = grep { ! /^\s*$/ } @data[0..15];

	if ( $onblocks !~ /^\d{2}$/ || $nblocks != $onblocks) {

	    warn "$file: $recn: $id: weird number of blocks: header says $onblocks. found $nblocks\n";
	    warn $record;
	}

        $tbl{$id}
          = [ map { { id => $data[$_], wait => $data[ $_ + 16 ] + 0 } } 0 .. $nblocks-1 ]
    }

    return \%tbl;
}



1;


__END__

=head1 NAME

CXC::ACIS::Command::OFLS - read OFLS SI mode configuration and raw command blocks

=head1 SYNOPSIS

  use CXC::ACIS::Command::OFLS qw[ loadOFLS_cfg loadOFLS_dat ];

  my $cfg = loadOFLS_cfg( 'current.cfg' );
  my $dat = loadOFLS_dat( 'current.dat' );

=head1 DESCRIPTION

This module provides functions for reading the ACIS OFLS SI mode
configuration file and the command blocks file.


=head1 FUNCTIONS

=head2 loadOFLS_cfg

  my $cfg = loadOFLS_cfg( $config_file );

Read in an OFLS ACIS SI mode configuration file.  It returns a
reference to a hash, which is keyed off of the SI mode.  The hash values are
references to arrays which identify the ACIS command blocks associated
with the SI mode.  Each element in the array is a hash with the
following fields:

=over

=item id

The command block id.

=item wait

The command block wait time.

=back

=head2 loadOFLS_dat

  my $dat = loadOFLS_dat( $command_block_file );

Read a raw ACIS OFLS command block file.  It returns a reference to a
hash, which is keyed off of the block id.  The hash values are the raw
commands.  Use L<B<CXC::ACIS::Command::Block>> to decode them.

=head1 LICENSE AND COPYRIGHT

Copyright (c) 2014 The Smithsonian Astrophysical Observatory

B<CXC::ACIS::Command::OFLS> is free software: you can redistribute it
and/or modify it under the terms of the GNU General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

=head1 AUTHOR

Diab Jerius  E<lt>djerius@cfa.harvard.eduE<gt>
